import os
import sys
from typing import Any

import docker  # type: ignore

from treebeard.buildtime.helper import run_image
from treebeard.conf import treebeard_env
from treebeard.util import fatal_exit


def run(project_id: str, notebook_id: str, run_id: str, image_name: str) -> int:
    client: Any = docker.from_env()  # type: ignore
    client.images.pull(image_name)
    return run_image(project_id, notebook_id, run_id, image_name)


if __name__ == "__main__":
    if not treebeard_env.notebook_id:
        raise Exception("No notebook ID at buildtime")
    if not treebeard_env.project_id:
        raise Exception("No project ID at buildtime")

    image_name_key = "TREEBEARD_IMAGE_NAME"
    image_name = os.getenv(image_name_key)
    if not image_name:
        fatal_exit(f"No image supplied under {image_name_key}")
    sys.exit(
        run(
            treebeard_env.project_id,
            treebeard_env.notebook_id,
            treebeard_env.run_id,
            image_name,
        )
    )



import glob
import re
import sys
from pydoc import ModuleScanner
from typing import Any, Optional, Set

import click
from nbconvert import ScriptExporter  # type: ignore
from sentry_sdk import capture_exception  # type: ignore

from_regexp = re.compile(r"^from (\w+)")
import_regexp = re.compile(r"^import (\w+)")


def get_imported_modules(glob_path: str) -> Set[str]:
    nb_paths = glob.glob(glob_path, recursive=True)
    if len(nb_paths) == 0:
        raise Exception(f"No notebooks found with glob {glob_path}")

    se: Any = ScriptExporter()

    all_imported_modules = set()
    for path in nb_paths:
        try:
            click.echo(f"Import check inspecting {path}")
            [script, _] = se.from_filename(path)
        except Exception as ex:
            capture_exception(ex)
            click.echo(f"Import checker cannot read {path}")
            continue
        lines = script.split("\n")

        for line in lines:
            from_match = from_regexp.match(line)
            if from_match:
                all_imported_modules.add(from_match.group(1))
                continue

            import_match = import_regexp.match(line)
            if import_match:
                all_imported_modules.add(import_match.group(1))
                continue

    return all_imported_modules


def get_installed_modules() -> Set[str]:
    modules = set()

    def callback(path: Optional[str], modname: str, desc: str):
        if modname and modname.endswith(".__init__"):
            modname = modname.replace(".__init__", " (package)")

        if modname.find(".") == -1:
            modules.add(modname)

    def onerror(modname: str):
        pass

    ModuleScanner().run(callback, onerror=onerror)
    return modules


def is_local_module(module_name: str):
    # is there a <name>.py or <name> directory
    dirs = glob.glob(f"**/{module_name}", recursive=True)
    scripts = glob.glob(f"**/{module_name}.py", recursive=True)

    return dirs or scripts


def check_imports(glob_path: str = "**/*ipynb"):
    click.echo(f"🌲 Checking for potentially missing imports...\n")
    imported_modules = get_imported_modules(glob_path)
    installed_modules = get_installed_modules()

    missing_modules = list(
        filter(
            lambda m: not is_local_module(m),
            imported_modules.difference(installed_modules),
        )
    )

    if len(missing_modules) > 0:
        click.echo(
            f"\n❗📦 You *may* be missing project requirements, the following modules are imported from your notebooks but can't be imported from your project root directory\n"
        )
        for module in sorted(missing_modules):
            click.echo(f"  - {module}")
        return False
    else:
        click.echo(
            "\n✨ All notebook imports appear to be backed by a reproducible dependency file!"
        )
        return True


if __name__ == "__main__":
    glob_path = sys.argv[1] if len(sys.argv) > 1 else "**/*ipynb"
    check_imports(glob_path)



from typing import IO, Any, List

import click
import requests

from treebeard.conf import (
    secrets_endpoint,
    treebeard_config,
    treebeard_env,
    validate_notebook_directory,
)
from treebeard.secrets.helper import get_secrets_archive


@click.group()
def secrets():
    """Store credentials used by your notebook"""
    pass


@secrets.command()  # type: ignore
@click.argument("files", type=click.File("r"), nargs=-1)
@click.option("--confirm/--no-confirm", default=False)
def push(files: List[IO[Any]], confirm: bool):
    push_secrets(files, confirm)


def push_secrets(files: List[IO[Any]], confirm: bool):
    """Uploads files marked in treebeard.yaml as 'secret'"""
    validate_notebook_directory(treebeard_env, treebeard_config)
    click.echo(
        f"🌲 Pushing secrets for {treebeard_env.project_id}/{treebeard_env.notebook_id}"
    )
    secrets_archive = get_secrets_archive(files, confirm=confirm)
    response = requests.post(  # type: ignore
        secrets_endpoint,
        files={"secrets": open(secrets_archive.name, "rb")},
        headers=treebeard_env.dict(),
    )
    if response.status_code != 200:
        click.echo(
            f"Error: service failure pushing secrets, {response.status_code}: {response.text}"
        )
        return

    click.echo("🔐  secrets pushed\n")



import configparser
import os
import time
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import List, Optional

import click
import yaml
from pydantic import BaseModel, ValidationError, validator  # type: ignore

from treebeard.util import fatal_exit


class TreebeardEnv(BaseModel):
    notebook_id: Optional[
        str
    ] = None  # Not present when CLI is not in notebook directory
    project_id: Optional[str] = None  # Not present when initially installing
    run_id: str
    email: Optional[str] = None  # Not present at build time
    api_key: Optional[str] = None  # Not present at build time

    def __str__(self) -> str:
        dict_obj = self.dict()
        if self.api_key:
            unredacted_length = 4
            dict_obj["api_key"] = (
                "x" * (len(self.api_key) - unredacted_length)
                + self.api_key[-unredacted_length:]
            )
        return str(dict_obj)


class TreebeardConfig(BaseModel):
    notebooks: List[str] = ["**/*.ipynb"]
    output_dirs: List[str] = ["output"]
    ignore: List[str] = []
    secret: List[str] = []
    kernel_name: str = "python3"
    strict_mode: bool = True
    cell_execution_timeout_seconds: int = 300
    schedule: Optional[str] = None

    @validator("schedule")
    def schedule_valdiator(cls, v: Optional[str]):
        schedules = (None, "hourly", "daily", "weekly")
        if v not in schedules:
            raise ValueError(f"schedule must be one of {schedules}")

        return v and v.lower()

    def get_deglobbed_notebooks(self):
        # warning: sensitive to current directory
        deglobbed_notebooks = []
        for pattern in self.notebooks:
            deglobbed_notebooks.extend(sorted(glob(pattern, recursive=True)))
        if len(deglobbed_notebooks) == 0:
            raise Exception(
                f"No notebooks found in project! Searched for {self.notebooks}"
            )

        ignored_notebooks = []
        for pattern in self.ignore:
            ignored_notebooks.extend(sorted(glob(pattern, recursive=True)))

        return [nb for nb in deglobbed_notebooks if nb not in ignored_notebooks]


env = "production"
if os.getenv("TREEBEARD_ENVIRONMENT"):
    env = os.getenv("TREEBEARD_ENVIRONMENT")


if env == "development":
    click.echo("WARNING: RUNNING IN LOCAL MODE", err=True)
    api_url = "http://localhost:8080"
    treebeard_web_url = "https://localhost:8000"
else:
    api_url = "https://api.treebeard.io"
    treebeard_web_url = "https://treebeard.io"


def get_run_path(treebeard_env: TreebeardEnv):
    return (
        f"{treebeard_env.project_id}/{treebeard_env.notebook_id}/{treebeard_env.run_id}"
    )


def get_time():
    return datetime.now().strftime("%H:%M:%S")


def get_config_path():
    home = str(Path.home())
    return f"{home}/.treebeard"


def validate_notebook_directory(
    treebeard_env: TreebeardEnv, treebeard_config: TreebeardConfig
):
    if treebeard_env.project_id is None:
        fatal_exit("No account config detected! Please run `treebeard configure`")

    if not Path("treebeard.yaml").is_file():
        click.secho(  # type: ignore
            "Warning: treebeard.yaml file not found! `treebeard setup` fetches an example.",
            fg="yellow",
        )
        click.echo("Using defaults: notebooks: - '**/*.ipynb' output_dirs: - 'outputs'")

    notebook_files = treebeard_config.get_deglobbed_notebooks()
    if not notebook_files:
        fatal_exit("No notebooks found in project! Treebeard expects at least one.")

    for notebook in notebook_files:
        if not os.path.exists(notebook):
            fatal_exit(
                f"Cannot run non-existent notebook '{notebook}', you should be in a project directory with a treebeard.yaml file"
            )


def get_treebeard_config() -> TreebeardConfig:
    notebook_config = "treebeard.yaml"
    if not os.path.exists(notebook_config):
        return TreebeardConfig()

    with open(notebook_config) as f:
        conf = yaml.load(f, Loader=yaml.FullLoader)  # type: ignore
        if not conf:
            return TreebeardConfig()
        try:
            return TreebeardConfig(**conf)
        except ValidationError as e:  # type: ignore
            fatal_exit(f"Error parsing treebeard.yaml\n{e.json()}")  # type: ignore


def get_treebeard_env():
    """Reads variables from a local file, credentials.cfg"""
    treebeard_project_id = os.getenv("TREEBEARD_PROJECT_ID")
    run_id = os.getenv("TREEBEARD_RUN_ID")
    if run_id is None:
        run_id = f"local-{int(time.time())}"

    notebook_id = os.getenv("TREEBEARD_NOTEBOOK_ID")
    if not notebook_id:
        notebook_id = Path(os.getcwd()).name

    email = None
    api_key = None

    # .treebeard config is present in CLI in place of env variables
    if os.path.exists(config_path):
        config = configparser.RawConfigParser()
        config.read(config_path)
        email = config.get("credentials", "treebeard_email")
        treebeard_project_id = config.get("credentials", "treebeard_project_id")
        api_key = config.get("credentials", "treebeard_api_key")

    return TreebeardEnv(
        notebook_id=notebook_id,
        project_id=treebeard_project_id,
        run_id=run_id,
        email=email,
        api_key=api_key,
    )


config_path = get_config_path()
treebeard_config = get_treebeard_config()
treebeard_env = get_treebeard_env()
run_path = get_run_path(treebeard_env)
secrets_endpoint = (
    f"{api_url}/{treebeard_env.project_id}/{treebeard_env.notebook_id}/secrets"
)
runner_endpoint = (
    f"{api_url}/{treebeard_env.project_id}/{treebeard_env.notebook_id}/runs"
)
service_status_endpoint = f"{api_url}/service_status"



from datetime import datetime
from typing import Any, Optional

import magic  # type: ignore
import requests
from requests import Response

from treebeard.conf import api_url

mime: Any = magic.Magic(mime=True)  # type: ignore


def log(message: str):
    print(f'{datetime.now().strftime("%H:%M:%S")}: {message}')


def upload_artifact(
    filename: str,
    upload_path: str,
    status: Optional[str],
    set_as_thumbnail: bool = False,
):
    log(f"Saving {filename} to {upload_path}")
    content_type: str = mime.from_file(filename)

    get_url_params = {"content_type": content_type}
    put_object_headers = {"Content-Type": content_type}
    if status:
        get_url_params["status"] = status
        put_object_headers["x-goog-meta-status"] = status

    with open(filename, "rb") as data:
        resp: Response = requests.get(  # type: ignore
            f"{api_url}/get_upload_url/{upload_path}", params=get_url_params,
        )
        if resp.status_code != 200:
            raise (
                Exception(
                    f"Get signed url failed for {filename}, {resp.status_code}\n{resp.text}"
                )
            )
        signed_url: str = resp.text
        put_resp = requests.put(signed_url, data, headers=put_object_headers,)  # type: ignore
        if put_resp.status_code != 200:
            raise (
                Exception(
                    f"Put object failed for {filename}, {put_resp.status_code}\n{put_resp.text}"
                )
            )

        qs = "set_as_thumbnail=true" if set_as_thumbnail else ""
        requests.post(f"{api_url}/{upload_path}/create_extras?{qs}")  # type: ignore



import sentry_sdk  # type: ignore
from sentry_sdk import configure_scope  # type: ignore

from treebeard.conf import env, treebeard_env
from treebeard.version import get_version


def setup_sentry():
    sentry_sdk.init(  # type: ignore
        "https://58543632a309471a88bb99f4f6bbdca0@sentry.io/2846147", environment=env
    )

    with configure_scope() as scope:  # type: ignore
        scope.set_tag("treebeard_version", get_version())  # type: ignore
        scope.set_tag("treebeard_project_id", str(treebeard_env.project_id))  # type: ignore
        scope.set_tag("treebeard_notebook_id", str(treebeard_env.notebook_id))  # type: ignore



from typing import Any, Dict, List

from pydantic import BaseModel


class Artifact(BaseModel):
    content_type: str
    name: str
    size: str


class Run(BaseModel):
    start_time: str
    end_time: str = ""
    run_id: str
    artifacts: List[Artifact]
    status: str
    trigger: Dict[str, Any] = {"mechanism": ""}
    url: str



import os
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from traceback import format_exc
from typing import Dict

import click
import papermill as pm  # type: ignore
from sentry_sdk import capture_exception  # type: ignore

from treebeard.conf import run_path, treebeard_config, treebeard_env
from treebeard.importchecker.imports import check_imports
from treebeard.logs.helpers import clean_log_file
from treebeard.runtime.helper import log, upload_artifact

bucket_name = "treebeard-notebook-outputs"

notebook_files = treebeard_config.get_deglobbed_notebooks()

notebook_status_descriptions = {
    "✅": "SUCCESS",
    "⏳": "WORKING",
    "❌": "FAILURE",
}


def save_artifacts(notebook_statuses: Dict[str, str]):
    with ThreadPoolExecutor(max_workers=4) as executor:
        log(f"Uploading outputs...")

        if treebeard_config is None:
            raise Exception("No Treebeard Config Present at runtime!")

        notebooks_files = treebeard_config.get_deglobbed_notebooks()
        first = True
        for notebook_path in notebooks_files:
            notebook_upload_path = f"{run_path}/{notebook_path}"
            executor.submit(
                upload_artifact,
                notebook_path,
                notebook_upload_path,
                notebook_status_descriptions[notebook_statuses[notebook_path]],
                set_as_thumbnail=first,
            )
            first = False

        for output_dir in treebeard_config.output_dirs:
            for root, _, files in os.walk(output_dir, topdown=False):
                for name in files:
                    full_name = os.path.join(root, name)
                    upload_path = f"{run_path}/{full_name}"
                    executor.submit(upload_artifact, full_name, upload_path, None)

        if os.path.exists("treebeard.log"):
            executor.submit(
                upload_artifact, "treebeard.log", f"{run_path}/treebeard.log", None
            )


def run_notebook(notebook_path: str) -> str:
    try:
        notebook_dir, notebook_name = os.path.split(notebook_path)
        log(
            f"Executing Notebook {notebook_name} in {'.' if len(notebook_dir) == 0 else notebook_dir}"
        )
        pm.execute_notebook(  # type: ignore
            notebook_path,
            notebook_path,
            kernel_name=treebeard_config.kernel_name,
            progress_bar=False,
            request_save_on_cell_execute=True,
            autosave_cell_every=10,
            execution_timeout=treebeard_config.cell_execution_timeout_seconds,
            log_output=True,
            cwd=f"{os.getcwd()}/{notebook_dir}",
        )
        log(f"✅ Notebook {notebook_path} passed!\n")
        return "✅"
    except Exception:
        tb = format_exc()
        log(f"❌ Notebook {notebook_path} failed!\n\n{tb}")
        return "❌"


def run(project_id: str, notebook_id: str, run_id: str) -> Dict[str, str]:
    log(f"🌲 treebeard runtime: running repo")
    subprocess.run(
        [
            "bash",
            "-c",
            """
            echo "working directory is $(pwd)\n\n$(ls -la)\n"
            """,
        ]
    )

    if treebeard_config is None:
        raise Exception("No Treebeard Config Present at runtime!")

    for output_dir in treebeard_config.output_dirs:
        os.makedirs(output_dir, exist_ok=True)

    notebook_statuses = {notebook: "⏳" for notebook in notebook_files}
    print(f"Will run the following:")
    [print(nb) for nb in notebook_files]
    print()

    for i, notebook_path in enumerate(notebook_files):
        log(f"⏳ Running {i + 1}/{len(notebook_files)}: {notebook_path}")
        notebook_statuses[notebook_path] = run_notebook(notebook_path)

    save_artifacts(notebook_statuses)
    log("Run Finished\n")

    return notebook_statuses


def start():
    if not treebeard_env.notebook_id:
        raise Exception("No notebook ID at runtime")
    if not treebeard_env.project_id:
        raise Exception("No project ID at buildtime")

    clean_log_file()

    notebook_statuses = run(
        treebeard_env.project_id, treebeard_env.notebook_id, treebeard_env.run_id
    )

    for notebook in notebook_statuses.keys():
        print(f"{notebook_statuses[notebook]} {notebook}")
    print()

    n_failed = len(list(filter(lambda v: v != "✅", notebook_statuses.values())))

    if n_failed > 0:
        log(f"{n_failed} of {len(notebook_statuses)} notebooks failed.\n")

        try:
            if treebeard_config.kernel_name == "python3":
                result = check_imports()

                if treebeard_config.strict_mode:
                    click.echo(
                        f"\nℹ️ If you would like to ignore notebook run failures and only fail on missing dependencies, add `strict_mode: False` to a `treebeard.yaml` file"
                    )
                else:
                    if result:
                        click.echo(
                            f"\nℹ️ Strict mode is disabled and import checker passed, run is successful! ✅"
                        )
                        sys.exit(0)
        except Exception as ex:
            click.echo(f"Import checker encountered and error...")
            capture_exception(ex)
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    start()



import warnings
from typing import Any

import click
from halo import Halo  # type: ignore
from humanfriendly import format_size, parse_size  # type: ignore
from timeago import format as timeago_format  # type: ignore

from treebeard.conf import treebeard_env
from treebeard.helper import CliContext, check_for_updates
from treebeard.notebooks.commands import cancel, run, status
from treebeard.other.commands import config, configure, setup, version
from treebeard.secrets.commands import secrets
from treebeard.sentry_setup import setup_sentry

project_id = treebeard_env.project_id
notebook_id = treebeard_env.notebook_id

warnings.filterwarnings(
    "ignore", "Your application has authenticated using end user credentials"
)


@click.group()
@click.option("--debug/--no-debug", default=False)
@click.pass_context  # type: ignore
def cli(ctx: Any, debug: bool):
    """🌲 
      
      code: https://github.com/treebeardtech/treebeard
      
      docs: https://treebeard.readthedocs.io/
    """
    setup_sentry()
    ctx.obj = CliContext(debug=debug)
    pass


@cli.resultcallback()
def process_result(*args: Any, **kwargs: Any):
    check_for_updates()


cli.add_command(configure)  # type: ignore
cli.add_command(status)  # type: ignore
cli.add_command(run)  # type: ignore
cli.add_command(cancel)  # type: ignore
cli.add_command(secrets)  # type: ignore
cli.add_command(setup)  # type: ignore
cli.add_command(config)  # type: ignore
cli.add_command(version)  # type: ignore



import datetime
import glob
import json
import os
import os.path
import pprint
import shutil
import subprocess
import sys
import tarfile
import tempfile
import time
from distutils.dir_util import copy_tree
from typing import Any, List

import click
import docker  # type: ignore
import requests
import yaml
from dateutil import parser
from halo import Halo  # type: ignore
from humanfriendly import format_size, parse_size  # type: ignore
from timeago import format as timeago_format  # type: ignore

from treebeard.buildtime.run_repo import run_repo
from treebeard.conf import (
    api_url,
    config_path,
    get_time,
    runner_endpoint,
    treebeard_config,
    treebeard_env,
    validate_notebook_directory,
)
from treebeard.helper import CliContext, sanitise_notebook_id
from treebeard.notebooks.types import Run
from treebeard.secrets.commands import push_secrets as push_secrets_to_store
from treebeard.secrets.helper import get_secrets_archive
from treebeard.util import fatal_exit

pp = pprint.PrettyPrinter(indent=2)

notebook_id = treebeard_env.notebook_id
project_id = treebeard_env.project_id


@click.command()
@click.option(
    "watch", "--watch", is_flag=True, help="Run and check completed build status"
)
@click.option(
    "local", "--local", is_flag=True, help="Build image with local docker installation",
)
@click.option("-n", "--notebooks", help="Notebooks to be run", multiple=True)
@click.option(
    "-i",
    "--ignore",
    help="Don't submit unneeded virtual envs and large files",
    multiple=True,
)
@click.option(
    "--confirm/--no-confirm", default=False, help="Confirm all prompt options"
)
@click.option(
    "--push-secrets/--no-push-secrets",
    default=False,
    help="Confirm all prompt options except pushing secrets",
)
@click.option(
    "--dockerless/--no-dockerless",
    default=False,
    help="Run locally without docker container",
)
@click.pass_obj  # type: ignore
def run(
    cli_context: CliContext,
    watch: bool,
    notebooks: List[str],
    ignore: List[str],
    local: bool,
    confirm: bool,
    push_secrets: bool,
    dockerless: bool,
):
    """
    Run a notebook and optionally schedule it to run periodically
    """
    notebooks = list(notebooks)
    ignore = list(ignore)

    validate_notebook_directory(treebeard_env, treebeard_config)

    # Apply cli config overrides
    treebeard_yaml_path: str = tempfile.mktemp()  # type: ignore
    with open(treebeard_yaml_path, "w") as yaml_file:
        if notebooks:
            treebeard_config.notebooks = notebooks

        yaml.dump(treebeard_config.dict(), yaml_file)  # type: ignore

    if dockerless:
        click.echo(
            f"🌲  Running locally without docker using your current python environment"
        )
        if not confirm and not click.confirm(
            f"Warning: This will clear the outputs of your notebooks, continue?",
            default=True,
        ):
            sys.exit(0)

        # Note: import runtime.run causes win/darwin devices missing magic to fail at start
        import treebeard.runtime.run

        treebeard.runtime.run.start()  # will sys.exit

    params = {}
    if treebeard_config.schedule:
        if confirm or click.confirm(
            f"📅 treebeard.yaml contains schedule '{treebeard_config.schedule}'. Enable it?"
        ):
            params["schedule"] = treebeard_config.schedule

    if (
        not local
        and len(treebeard_config.secret) > 0
        and not confirm
        and not push_secrets
    ):
        push_secrets = click.confirm("Push secrets first?", default=True)

    if push_secrets:
        push_secrets_to_store([], confirm=confirm)

    if treebeard_config:
        ignore += (
            treebeard_config.ignore
            + treebeard_config.secret
            + treebeard_config.output_dirs
        )

    click.echo("🌲  Copying project to tempdir and stripping notebooks")

    temp_dir = tempfile.mkdtemp()
    copy_tree(os.getcwd(), str(temp_dir), preserve_symlinks=1)
    notebooks_files = treebeard_config.get_deglobbed_notebooks()
    for notebooks_file in notebooks_files:
        try:
            subprocess.check_output(["nbstripout"] + notebooks_file, cwd=temp_dir)
        except:
            print(f"Failed to nbstripout {notebooks_file}! Is it valid?")
    click.echo(notebooks_files)
    click.echo("🌲  Compressing Repo")

    with tempfile.NamedTemporaryFile(
        "wb", suffix=".tar.gz", delete=False
    ) as src_archive:
        with tarfile.open(fileobj=src_archive, mode="w:gz") as tar:

            def zip_filter(info: tarfile.TarInfo):
                if info.name.endswith("treebeard.yaml"):
                    return None

                for ignored in ignore:
                    if info.name in glob.glob(ignored, recursive=True):
                        return None

                # if len(git_files) > 0 and info.name not in git_files:
                #     return None
                click.echo(f"  Including {info.name}")
                return info

            tar.add(
                str(temp_dir), arcname=os.path.basename(os.path.sep), filter=zip_filter,
            )
            tar.add(config_path, arcname=os.path.basename(config_path))
            tar.add(treebeard_yaml_path, arcname="treebeard.yaml")

    if not confirm and not click.confirm(
        "Confirm source file set is correct?", default=True
    ):
        click.echo("Exiting")
        sys.exit()

    if local:
        build_tag = str(time.mktime(datetime.datetime.today().timetuple()))
        repo_image_name = f"gcr.io/treebeard-259315/projects/{project_id}/{sanitise_notebook_id(str(notebook_id))}:{build_tag}"
        click.echo(f"🌲  Building {repo_image_name} Locally\n")
        secrets_archive = get_secrets_archive()
        repo_url = f"file://{src_archive.name}"
        secrets_url = f"file://{secrets_archive.name}"
        status = run_repo(
            str(project_id),
            str(notebook_id),
            treebeard_env.run_id,
            build_tag,
            repo_url,
            secrets_url,
            branch="cli",
            local=True,
        )
        click.echo(f"Local build exited with status code {status}")
        sys.exit(status)

    size = os.path.getsize(src_archive.name)
    max_upload_size = "100MB"
    if size > parse_size(max_upload_size):
        fatal_exit(
            click.style(
                (
                    f"ERROR: Compressed notebook directory is {format_size(size)},"
                    f" max upload size is {max_upload_size}. \nPlease ensure you ignore any virtualenv subdirectory"
                    " using `treebeard run --ignore venv`"
                ),
                fg="red",
            )
        )

    time_seconds = int(time.mktime(datetime.datetime.today().timetuple()))
    build_tag = str(time_seconds)

    upload_api = f"{api_url}/source_upload_url/{project_id}/{notebook_id}/{build_tag}"
    resp = requests.get(upload_api)  # type: ignore

    signed_url: str = resp.text
    put_resp = requests.put(  # type: ignore
        signed_url,
        open(src_archive.name, "rb"),
        headers={"Content-Type": "application/x-tar"},
    )
    assert put_resp.status_code == 200

    if os.getenv("GITHUB_ACTIONS"):
        params["event"] = os.getenv("GITHUB_EVENT_NAME")
        params["sha"] = os.getenv("GITHUB_SHA")
        params["branch"] = os.getenv("GITHUB_REF").split("/")[-1]
        workflow = os.getenv("GITHUB_WORKFLOW")
        params["workflow"] = (
            workflow.replace(".yml", "").replace(".yaml", "").split("/")[-1]
        )

    click.echo(f"🌲  submitting archive to runner ({format_size(size)})...")
    submit_endpoint = f"{api_url}/runs/{treebeard_env.project_id}/{treebeard_env.notebook_id}/{build_tag}"
    response = requests.post(  # type: ignore
        submit_endpoint,
        params=params,
        headers={"api_key": treebeard_env.api_key, "email": treebeard_env.email},
    )
    shutil.rmtree(temp_dir)

    if response.status_code != 200:
        raise click.ClickException(f"Request failed: {response.text}")

    try:
        json_data = json.loads(response.text)
        click.echo(f"✨  Run has been accepted! {json_data['admin_url']}")
    except:
        click.echo("❗  Request to run failed")
        click.echo(sys.exc_info())

    if watch:
        build_result = None
        while not build_result:
            time.sleep(5)
            response = requests.get(runner_endpoint, headers=treebeard_env.dict())  # type: ignore
            json_data = json.loads(response.text)
            if len(json_data["runs"]) == 0:
                status = "FAILURE"
            else:
                status = json_data["runs"][-1]["status"]
            click.echo(f"{get_time()} Build status: {status}")
            if status == "SUCCESS":
                build_result = status
                click.echo(f"Build result: {build_result}")
            elif status in ["FAILURE", "TIMEOUT", "INTERNAL_ERROR", "CANCELLED"]:
                fatal_exit(f"Build failed")


@click.command()
def cancel():
    """Cancels the current notebook build and schedule"""
    validate_notebook_directory(treebeard_env, treebeard_config)
    notif = f"🌲  Cancelling {notebook_id}"
    spinner: Any = Halo(text=notif, spinner="dots")
    spinner.start()
    requests.delete(runner_endpoint, headers=treebeard_env.dict())  # type: ignore
    spinner.stop()
    click.echo(f"{notif}...🛑 cancellation confirmed!")


@click.command()
def status():
    """Show the status of the current notebook"""
    validate_notebook_directory(treebeard_env, treebeard_config)
    response = requests.get(runner_endpoint, headers=treebeard_env.dict())  # type: ignore
    if response.status_code != 200:
        raise click.ClickException(f"Request failed: {response.text}")

    json_data = json.loads(response.text)
    if len(json_data) == 0:
        fatal_exit(
            "This notebook has not been run. Try running it with `treebeard run`"
        )
    click.echo("🌲  Recent runs:\n")

    max_results = 5
    status_emoji = {
        "SUCCESS": "✅",
        "QUEUED": "💤",
        "WORKING": "⏳",
        "FAILURE": "❌",
        "TIMEOUT": "⏰",
        "CANCELLED": "🛑",
    }

    runs: List[Run] = [Run.parse_obj(run) for run in json_data["runs"][-max_results:]]  # type: ignore
    for run in runs:
        now = parser.isoparse(datetime.datetime.utcnow().isoformat() + "Z")
        start_time = parser.isoparse(run.start_time)
        time_string: str = timeago_format(start_time, now=now)

        mechanism: str = run.trigger["mechanism"]
        ran_via = "" if len(mechanism) == 0 else f"via {mechanism}"
        try:
            branch = f"🔀{run.trigger['branch']}"
        except:
            branch = ""

        click.echo(
            f"  {status_emoji[run.status]}  {time_string} {ran_via} {branch} -- {run.url}"
        )



import pprint
import warnings
from pathlib import Path

import click
from halo import Halo  # type: ignore
from humanfriendly import format_size, parse_size  # type: ignore
from timeago import format as timeago_format  # type: ignore

from treebeard.conf import treebeard_env
from treebeard.helper import create_example_yaml, set_credentials
from treebeard.util import fatal_exit
from treebeard.version import get_version

pp = pprint.PrettyPrinter(indent=2)


notebook_id = treebeard_env.notebook_id

warnings.filterwarnings(
    "ignore", "Your application has authenticated using end user credentials"
)


@click.command()
@click.option("--email")
@click.option("--api_key")
@click.option("--project_id")
def configure(email: str, api_key: str, project_id: str):
    """Register with Treebeard services"""
    set_credentials(email, api_key, project_id)


@click.command()
def setup():
    """Creates examples treebeard.yaml configuration file"""
    if Path("treebeard.yaml").is_file():
        fatal_exit("📁 found existing treebeard.yaml file here")
    create_example_yaml()
    click.echo("📁 created example treebeard.yaml, please update it for your project")


@click.command()
def version():
    """Shows treebeard package version"""
    click.echo(get_version())


@click.group()
def config():
    """Shows Treebeard internal configuration"""


@config.command()  # type: ignore
def list():
    click.echo(pp.pformat(treebeard_env.dict()))


@config.command()  # type: ignore
@click.argument("key", type=click.STRING)
def get(key: str):
    if key in treebeard_env.dict():
        click.echo(treebeard_env.dict()[key])
    else:
        click.echo(f"There is no value for {key}")



def log(content: str):
    """Manually log your custom content to a local file
    The file is saved at the end of the run and uploaded as an output
    then fetched by the notification function
    Logged content will be presented in the Treebeard admin page,
    and sent via Slack

    Params: content, custom string used in a command from a notebook
    """
    with open("treebeard.log", "a+") as f:
        f.write("\n")
        f.write(content)



import os
import subprocess
import sys
from pathlib import Path
from traceback import format_exc
from typing import Any, Optional

import click
import docker  # type: ignore
from docker.errors import ImageNotFound, NotFound  # type: ignore

from treebeard.buildtime.helper import run_image
from treebeard.conf import get_treebeard_config, run_path, treebeard_env
from treebeard.helper import sanitise_notebook_id
from treebeard.util import fatal_exit


def download_archive(unzip_location: str, download_location: str, url: str):
    subprocess.check_output(
        [
            "bash",
            "-c",
            f'curl -o {download_location} "{url}" && tar -C {unzip_location} -xvf {download_location}',
        ]
    )


def fetch_image_for_cache(client: Any, image_name: str):
    try:
        click.echo(f"Pulling {image_name}")
        client.images.pull(image_name)
    except:
        click.echo(f"Could not pull image for cache, continuing without.")


def run_repo(
    project_id: str,
    notebook_id: str,
    run_id: str,
    build_tag: str,
    repo_url: str,
    secrets_url: Optional[str],
    branch: str,
    local: bool = False,
) -> int:
    click.echo(f"🌲 Treebeard buildtime, building repo")
    click.echo(f"Run path: {run_path}")

    client: Any = docker.from_env()  # type: ignore

    try:
        # Create bundle directory
        abs_notebook_dir = f"/tmp/{notebook_id}"
        Path(abs_notebook_dir).mkdir(parents=True, exist_ok=True)
        os.chdir(abs_notebook_dir)

        # Add repo to bundle
        download_archive(abs_notebook_dir, f"/tmp/{notebook_id}_repo.tgz", repo_url)

        if secrets_url:
            download_archive(
                abs_notebook_dir, f"/tmp/{notebook_id}_secrets.tgz", secrets_url
            )

        notebook_files = get_treebeard_config().get_deglobbed_notebooks()
        if len(notebook_files) == 0:
            raise Exception(
                "No notebooks found to run. If you are using a treebeard.yaml file, check it is correct: https://treebeard.readthedocs.io/en/latest/project_config.html"
            )
        try:
            subprocess.check_output(["nbstripout"] + notebook_files)
        except:
            print(
                f"Failed to nbstripout a notebook! Do you have an invalid .ipynb?\nNotebooks: {notebook_files}"
            )
    finally:
        click.echo("Treebeard Bundle Contents:")
        subprocess.run(["pwd"])
        subprocess.run(["ls", "-la", abs_notebook_dir])

    # Pull down images to use in cache
    image_name = f"gcr.io/treebeard-259315/projects/{sanitise_notebook_id(project_id)}/{sanitise_notebook_id(notebook_id)}"

    # Build image but don't run
    versioned_image_name = f"{image_name}:{build_tag}"
    passing_image_name = f"{image_name}:{branch}"
    latest_image_name = f"{image_name}:{branch}-latest"

    if not local:
        fetch_image_for_cache(client, latest_image_name)

    user_name = "project_user"  # All images having the same home dir enables caching
    r2d = f"""
    repo2docker \
        --debug \
        --no-run \
        --user-name {user_name} \
        --user-id 1000 \
        --image-name {versioned_image_name} \
        --cache-from {latest_image_name} \
        --target-repo-dir {abs_notebook_dir} \
        {abs_notebook_dir}
    """

    try:
        subprocess.check_output(["bash", "-c", r2d])
        click.echo(f"✨  Successfully built {versioned_image_name}")
    except:
        click.echo(f"\n\n❗ Failed to build container from the source repo")
        return 1

    subprocess.check_output(["docker", "tag", versioned_image_name, latest_image_name])
    if not local:
        try:
            click.echo(
                f"Image built: Pushing {versioned_image_name} and {latest_image_name}"
            )
            client.images.push(versioned_image_name)
            client.images.push(latest_image_name)  # this tag is necessary for caching
        except Exception:
            click.echo(
                f"Failed to push image, will try again on success\n{format_exc()}"
            )

    click.echo(f"Image built successfully, now running.")
    status = run_image(project_id, notebook_id, run_id, versioned_image_name)
    if status != 0:
        click.echo(f"Image run failed, not updated {passing_image_name}")
        return status

    subprocess.check_output(["docker", "tag", versioned_image_name, passing_image_name])
    click.echo(f"tagged {versioned_image_name} as {passing_image_name}")
    return 0


if __name__ == "__main__":
    build_tag_key = "TREEBEARD_BUILD_TAG"
    repo_url_key = "TREEBEARD_REPO_URL"
    secrets_url_key = "TREEBEARD_SECRETS_URL"
    branch_key = "TREEBEARD_BRANCH"

    subprocess.run(["bash", "-c", "echo Building repo"])

    build_tag = os.getenv(build_tag_key)
    if not build_tag:
        fatal_exit(f"No build_tag provided inside {build_tag_key}")

    branch = os.getenv(branch_key)
    if not branch:
        fatal_exit(f"No branch provided inside {branch_key}")

    repo_url = os.getenv(repo_url_key)
    if not repo_url:
        fatal_exit(f"No repo_url provided inside {repo_url_key}")

    secrets_url = os.getenv(secrets_url_key)
    if not treebeard_env.notebook_id:
        raise Exception("No notebook ID at runtime")
    if not treebeard_env.project_id:
        raise Exception("No project ID at buildtime")

    status = run_repo(
        treebeard_env.project_id,
        treebeard_env.notebook_id,
        treebeard_env.run_id,
        build_tag,
        repo_url,
        secrets_url,
        branch,
    )

    sys.exit(status)



import setuptools  # type: ignore

"""
Treebeard - CLI and Github app for running notebooks
====================================================

Check out [GitHub](https://github.com/treebeardtech/treebeard) to find out more.
"""

install_requires = [
    "click",
    "halo",
    "papermill",
    "jupyter-repo2docker",
    "colorama",
    "ipykernel",
    "pathlib",
    "timeago",
    "humanfriendly",
    "sentry-sdk",
    "pyyaml",
    "pydantic",
    "docker",
    "python-magic",
    "nbstripout",
    "docker",
    "requests",
    "jupyter-client>=6.1.0",  # fix Step #0: ERROR: nbclient 0.2.0 has requirement jupyter-client>=6.1.0, but you'll have jupyter-client 5.3.4 which is incompatible.
    "jupyter-contrib-nbextensions",  # for nbconvert scriptexporter
]

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("treebeard/version.txt", "r") as fh:
    version = fh.read()

setuptools.setup(  # type: ignore
    name="treebeard",
    version=version,
    author="Treebeard Technologies",
    author_email="alex@treebeard.io",
    description="Run and schedule jupyter notebook projects in the cloud",
    long_description=long_description,
    long_description_content_type="text/markdown",  # type: ignore
    packages=setuptools.find_namespace_packages(),  # type: ignore
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    license="Apache-2.0",
    python_requires=">=3.6",
    package_data={"treebeard": ["version.txt", "example_treebeard.yaml"]},
    entry_points={
        "console_scripts": [
            "treebeard = treebeard:cli",
            "treebeard-spaceship = treebeard.version:cli",
        ]
    },
    install_requires=install_requires,
    url="https://github.com/treebeardtech/treebeard",
)



import configparser
import json
import os
from pathlib import Path
from typing import Optional

import click
import requests
from pydantic import BaseModel

from treebeard.conf import config_path
from treebeard.version import get_version

version = get_version()


def set_credentials(email: str, key: str, project_id: str):
    """Create user credentials"""
    config = configparser.RawConfigParser()
    config.add_section("credentials")
    config.set("credentials", "TREEBEARD_EMAIL", email)
    config.set("credentials", "TREEBEARD_PROJECT_ID", project_id)
    config.set("credentials", "TREEBEARD_API_KEY", key)
    with open(config_path, "w") as configfile:
        config.write(configfile)
    click.echo(f"🔑  Config saved in {config_path}")
    return project_id


def check_for_updates():
    pypi_data = requests.get("https://pypi.org/pypi/treebeard/json")  # type: ignore
    latest_version = json.loads(pypi_data.text)["info"]["version"]

    if latest_version != version:
        click.echo(
            click.style(
                "🌲 Warning: you are not on the latest version of Treebeard, update with `pip install --upgrade treebeard`",
                fg="yellow",
            ),
            err=True,
        )


def get_service_status_message(service_status_url: str) -> Optional[str]:
    try:
        data = json.loads(requests.get(service_status_url).text)  # type: ignore
        if "message" in data:
            return data["message"]
    except:  # Non-200 status/timeout, etc.
        return None


class CliContext(BaseModel):
    debug: bool


def sanitise_notebook_id(notebook_id: str) -> str:
    out = notebook_id.replace("_", "-")
    out = out.replace(" ", "-")
    out = out.replace(".", "-")
    return out.lower()


def create_example_yaml():
    dirname = os.path.split(os.path.abspath(__file__))[0]
    with open(Path(f"{dirname}/example_treebeard.yaml"), "rb") as f:
        open("treebeard.yaml", "wb").write(f.read())
    return



from typing import Any

import click
import docker  # type: ignore


def run_image(project_id: str, notebook_id: str, run_id: str, image_name: str) -> int:
    client: Any = docker.from_env()  # type: ignore

    container = client.containers.run(
        image_name,
        "bash -c '(which treebeard > /dev/null || pip install treebeard) && python -m treebeard.runtime.run'",
        environment={
            "TREEBEARD_PROJECT_ID": project_id,
            "TREEBEARD_NOTEBOOK_ID": notebook_id,
            "TREEBEARD_RUN_ID": run_id,
        },
        detach=True,
    )

    [click.echo(line, nl=False) for line in container.logs(stream=True)]

    result = container.wait()
    return int(result["StatusCode"])
